(* $Id: godi_make.mli 569 2005-07-23 23:20:13Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann. 
 * 
 * This file is part of godi-tools. 
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Parsing and global evaluation of makefiles *)

module S : sig
  (* S.t is a string with already computed hash value *)
  include Hashtbl.HashedType
  val mk : string -> t
  val dest : t -> string
end


module H : Hashtbl.S with type key = S.t


type location = 
    { loc_filename : string;
      loc_linenumber : int;
    }

type make_cmd =
    M_assign of     (lhs * value)   (* VAR=VALUE *)
  | M_assign_app of (lhs * value)   (* VAR+=VALUE *)
  | M_assign_opt of (lhs * value)   (* VAR?=VALUE *)
  | M_assign_exp of (lhs * value)   (* VAR:=VALUE *)
  | M_assign_sh  of (lhs * value)   (* VAR!=VALUE *)
  | M_push of S.t
      (* Push the value of the variable on to the stack; undef the variable *)
  | M_pop
      (* Pop the variable from the stack *)
  | M_undef of S.t  (* .undef VAR *)
  | M_if of (expr * loc_make_cmd list * loc_make_cmd list) 
      (* .if <expr> .then <commands> .else <commands>
       * Also used for .ifdef, .ifndef, .ifmake, .ifnmake
       *)
  | M_for of (S.t list * value * loc_make_cmd list)
      (* .for <variables> in <value> \n <commands> \n .endfor
       *)
  | M_include of (bool * value)       (* .include "file" *)
  | M_sys_include of (bool * value)   (* .include <file> *)
      (* The bool says whether to ignore non-existing files *)
  | M_dep of (value * value)          (* TARGET : SOURCE *)
  | M_forced_dep of (value * value)   (* TARGET ! SOURCE *)
  | M_concat_dep of (value * value)   (* TARGET :: SOURCE *)
      (* Note: target and source are unexpanded *)
  | M_shell of value
      (* A shell command. It should be preceded by a M_*_dep command (after
       * resolving conditionals), but this is not checked by the parser.
       * The value can contain "@" and "-" at the beginning with the 
       * usual meaning.
       *)

and loc_make_cmd = location * make_cmd

and lhs =
    LHS_val of value      (* The name of the variable is the expanded value *)
  | LHS_hashed_var of S.t (* The name is the S.t component *)

and value =
  | V_lit of string
      (* An already expanded value *)
  | V_esc of char
      (* An escaped character (the character following a backslash)
       *)
  | V_var of varspec
      (* The value is the expansion of this variable *)
  | V_hashed_var of hashed_varspec
      (* The value is the expansion of this variable *)
  | V_replaced_var of replaced_varspec
      (* A variable that already has been replaced by a constant value *)
  | V_list of value list
      (* The value is the concatentation of the inner values *)

and expr =
    E_or of (expr list)               (* expr || expr ... *)
  | E_and of (expr list)              (* expr && expr ... *)
  | E_not of expr                     (* !expr *)
  | E_defined of value                (* defined(VAR) *)
  | E_make of value                   (* make(target) *)
  | E_exists of value                 (* exists(file) *)
  | E_empty of varspec                (* empty(varspec) *)
  | E_target of value                 (* target(target) *)
  | E_commands of value               (* commands(target) *)
  | E_num_eq of (value * value)       (* ${varspec} == number *)
  | E_num_ne of (value * value)       (* ${varspec} != number *)
  | E_num_le of (value * value)       (* ${varspec} <= number *)
  | E_num_lt of (value * value)       (* ${varspec} <  number *)
  | E_num_ge of (value * value)       (* ${varspec} >= number *)
  | E_num_gt of (value * value)       (* ${varspec} >  number *)
  | E_str_eq of (value * value)       (* ${varspec} == "string" *)
  | E_str_ne of (value * value)       (* ${varspec} != "string" *)

and varspec = value * modifier list   (* name, modifiers *)

and replaced_varspec = string * string * modifier list
                       (* name, value, modifiers *)

and hashed_varspec = S.t * modifier list

and modifier =
    MO_E
  | MO_H
  | MO_M of value
  | MO_N of value
  | MO_O
  | MO_Q
  | MO_R 
  | MO_tl
  | MO_tu
  | MO_S of (value * value * subst_modifier)
  | MO_C of (value * value * subst_modifier)
  | MO_T
  | MO_u
  | MO_if of (value * value)     (* ? str : str *)
  | MO_subst of (value * value)  (* str=str *)
  | MO_U of value
  | MO_D of value
  | MO_L
  | MO_P
  | MO_cmd of value              (* !cmd! *)
  | MO_sh                        (* sh *)
  | MO_map of (value * value)   (* @name@expr@ *)

  (* assignments inside varspecs are not supported *)

and subst_modifier = 
    { sm_first : bool;   (* Only first word is subject to substitution *)
      sm_global : bool;  (* All occurrences in the word/words are replaced *)
    }
;;


type magic_context ;;
  (* Intentionally hidden *)

type variable =
    V_normal of value
      (* A normal variable *)
  | V_env of value
      (* An environment variable *)
  | V_override of value
      (* A variable that cannot be modified because it is set by the
       * command call
       *)
  | V_undef
      (* The undefined value (so we can undefine variables that are contained
       * in the environment)
       *)
  | V_magic of (location -> magic_context -> value)
      (* The function is called to get the value. The magic_context is used
       * for variables like ${.TARGET} that change their value depending on
       * where they are expanded.
       * Magic variables cannot be modified.
       *)
;;


type rule =
    { rule_target : string;
      mutable rule_sources : string list;
        (* Files on the right hand of ":" *)
      mutable rule_atts : string list;
        (* Attributes are on the right hand of ":" and begin with a dot,
	 * e.g. .NOTMAIN
	 *)
      mutable rule_commands: (location * value) list;
      rule_type : [ `Dep | `Forced_dep | `Concat_dep ];
        (* `Dep: TARGET:SOURCE
         * `Forced_dep: TARGET!SOURCE
         * `Concat_dep: TARGET::SOURCE
         *)
      mutable rule_file_type : [ `File | `Phony | `Join ];
        (* `File: target is a file
         * `Phony: target is not a file and always remade (overriding rule_type)
         * `Join: target is not a file and commands are only executed if any
         *    sources are out of date (overriding rule_type)
         *)
      mutable rule_ignore_flag : bool;       (* .IGNORE *)
      mutable rule_made_flag : bool;         (* .MADE *)
      mutable rule_force_exec_flag : bool;   (* .MAKE *)
      mutable rule_optional_flag : bool;     (* .OPTIONAL *)
      mutable rule_silent_flag : bool;       (* .SILENT *)
      mutable rule_system_flag : bool;
              (* Whether the rule is defined in a file included using
               * the system search path, i.e. .include <file>.
               * (Internal: This flag is set when the variable .SYSTEM_SCOPE
               * is non-empty.)
               *)
    }
;;


type command =
    { command : string;
        (* The command string, to be executed by sh *)
      env : (string * string) list;
        (* Additional environment variables *)
      directory : string option;
        (* maybe change directory *)
    }
;;


type context =
    { vars : variable H.t;
        (* Current values of variables *)
      exported_vars : string list;
        (* Variables to export for command invocations. There is currently
         * no syntax to do that, but this feature is used for PWD and
         * MAKEFLAGS
         *)
      rules : rule list H.t;
        (* The rules, hashed by their targets. The following special
	 * targets are also recorded here (all other are dropped after
	 * they have been processed):
	 * .BEGIN, .END, .INTERRUPT
	 *)
      final_targets : string list;
        (* The targets to finally create (corresponds to the targets
	 * mentioned on the command line)
	 *)
      mutable default_target : string option;
	(* The default target that is made when there are no
	 * final_targets
	 *)
      mutable sys_mk_path : string list;
        (* The path where Makefiles are searched included with <...> notation *)
      mutable user_mk_path : string list;
        (* The path where Makefiles are searched included with "..." notation *)
      mutable curdir : string;
        (* The directory where [make] is invoked *)
      mutable objdir : string;
        (* The directory serving as virtual base directory. This MUST be
	 * an absolute path.
	 *)

      (* Note: Normally curdir = objdir. ${.CURDIR} is initialized from
       * curdir, ${.OBJDIR} is initialized from objdir. curdir and objdir
       * should point to the same directory. However, objdir is used
       * by "make" for its own file lookup, and counts.
       *)

      capture_sh_cmd : context -> command -> (string * Unix.process_status);
        (* Invoke an external command, return the output *)

      call_sh_cmd : context -> command -> Unix.process_status;
        (* Invoke an external command. It is unspecified what to do with
         * the output of the command
         *)

      mutable global_ignore_flag : bool;   (* .IGNORE for all rules *)
      mutable global_silent_flag : bool;   (* .SILENT for all rules *)
      mutable global_precious_flag : bool; (* .PRECIOUS for all targets *)

      mutable suffixes : string list;      (* list of suffixes *)

      mutable precious_targets : string list; (* list of .PRECIOUS targets *)

      stderr : Godi_exec.collector;
        (* Target for warnings etc. *)

      warn_on_env_access : bool;
        (* Whether to print a warning when an environment variable is
	 * accessed that is not listed in GODI_MAKE_ENV.
	 *)
    }
;;


exception Parse_error of (location * string);;

exception Eval_error of (location * string);;

exception Exec_error of (location * string);;

exception File_not_found of string;;

exception Target_not_found of string;;

val loc_unknown : location;;
  (* The unknown location: filename is "", line number 0 *)

val parse : ?pref_user_error:bool -> 
            ?system_scope:bool ->
            string -> 
              loc_make_cmd list
  (* Parses the file (but not included files), and returns the contents.
   * The returned list sets always the variables ".PARSEDIR" and 
   * ".PARSEFILE", and protects these variables with M_push and M_pop.
   *
   * The filename is NOT searched, neither it is turned into an absolute
   * pathname. Relative filenames are relative to cwd.
   *
   * pref_user_error: If set, the function raises User_error exceptions
   * instead of Parse_error and File_not_found (default: false)
   *
   * system_scope: If set, the parsed file is assumed to be a system file.
   * (Default: false)
   *)

val eval : 
      ?pref_user_error : bool ->
      ?include_user_file : (context -> string -> loc_make_cmd list) ->
      ?include_sys_file : (context -> string -> loc_make_cmd list) ->
      context -> loc_make_cmd list -> unit
  (* Evaluates the make commands for variable assignments, and modifies
   * the context.
   * ~include_user_file:
   *    This function is called to include a file. It may raise File_not_found
   *    to signal that the file does not exist. The default function
   *    searches the file (ctx.user_mk_path @ ctx.sys_mk_path) and parses it.
   * ~include_sys_file:
   *    This function is called to include a system file (using <...>
   *    notation). It may raise File_not_found to signal that the file does not
   *    exist. The default function searches the file (ctx.sys_mk_path)
   *    and parses it.
   *
   * pref_user_error: If set, the function raises User_error exceptions
   * instead of Eval_error and File_not_found (default: false)
   *)


val dump_context : ?env:bool -> context -> unit
  (* Debugging *)

val create_context :
      ?env:string array ->
      ?override:(string * string) list ->
      ?options:string ->
      ?user_mk_path:string list ->
      ?sys_mk_path:string list ->
      ?curdir:string ->
      ?objdir:string ->
      ?make_cmd:string ->
      ?capture_sh_cmd : (context -> command -> (string * Unix.process_status)) ->
      ?call_sh_cmd : (context -> command -> Unix.process_status) ->
      ?stderr:Godi_exec.collector ->
      ?warn_on_env_access:bool ->  (* default: false *)
      string list ->
	context
  (* Creates a context
   *
   * ~options:
   *    This string initializes .MAKEOPTIONS, the command-line options to
   *    pass down to recursive godi_make invocations
   *
   * ~run_sh_cmd:
   *    This function is called to read the output of an external command.
   *    The argument is the command to call. The returned pair (output,status) 
   *    is the output string of the command and the process status.
   *    The default function just starts the command.
   *
   * ~make_cmd:
   *    The name of the executable (for ${.MAKE}). Default is Sys.argv.(0)
   *)

val expand : 
      ?pref_user_error:bool -> 
      ?magic_context:magic_context -> 
      ?enable_env_check:bool ->
      context -> location -> value -> string
  (* Expand the value in the context, and return the fully expanded
   * result.
   * The [location] is only used for error messages.
   *
   * ~enable_env_check: Whether to print a warning when a variable is
   *    accessed that has not yet been initialized, and that is tried
   *    to be taken from the environment. The warning is not printed
   *    when the variable is listed in GODI_MAKE_ENV. Default:true.
   *)

val expand_var : 
      ?pref_user_error:bool -> 
      ?magic_context:magic_context -> 
      ?enable_env_check:bool ->
      context -> location -> string -> string
  (* Expand the value of the variable in the context, and return the 
   * fully expanded result.
   *
   * ~enable_env_check: see [expand]
   *)

val var_definition :
      context -> string -> string
  (* Return a string that can serve as the variable's definition *)


type parser_cache
  (* Caches parsed files *)

val create_parser_cache : unit -> parser_cache
  (* Creates an empty cache *)

val fast_parse_eval : 
      ?pref_user_error : bool ->
      ?sys_mk:bool -> 
      ?make_user_file:(string -> context -> unit) ->
      parser_cache -> string -> context -> unit
  (* The string argument is the file to parse. It is also evaluated,
   * hereby modifying the context. Included files are also parsed
   * and evaluated, using the search paths given in the context.
   * All parsed files are put into the parser_cache (together with
   * their mtimes). Files are extracted from the cache instead of
   * parsing them anew when it can be assumed that the files on the
   * disk have not changed.
   *
   * The filename argument may be relative. In this case, it is 
   * interpreted relative to ctx.objdir. It is not searched, however.
   *
   * sys_mk: If true (default), the contents of <sys.mk> are parsed
   *   before any other file.
   *
   * pref_user_error: Whether User_error exceptions are preferred over
   * Parse_error, Eval_error, and File_not_found. (Default: false)
   *
   * make_user_file: If passed, it is possible to create included files
   * by recursively invoking the executor. E.g. ".include file", and 
   * there is a rule to create "file". The string argument of the
   * function is the file name.
   *)

exception Out_of_date

val execute :
      ?pref_user_error : bool ->
      ?exec_mode:[ `Execute | `Simulate_except | `Simulate | `Test | `Fake ] ->
      context -> 
        unit
  (* Executes the context. There are several modes:
   *
   * - [`Execute]: Execute all commands. This is the default.
   * - [`Simulate_except]: Print the commands but do not execute them except
   *   the rule containing the commands is marked with .MAKE
   *   (command-line option -n)
   * - [`Simulate]:  Print the commands but do not execute them
   *   (command-line option -N)
   * - [`Test]: Only test whether the targets are up to date. Raises the
   *   exception Out_of_date if a target is not up to date. 
   *   (Command-line option -q.)
   * - [`Fake]: Instead of executing the commands of a rule, only create
   *   an empty file or update the mtime (command-line option -t)
   *
   * pref_user_error: If set, the function raises User_error exceptions
   * instead of XXX (default: false)
   *)


(* Limitations of this "make" implementation: 
 * See trailing comment in godi_make.ml 
 *)
